# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2022 NITK Surathkal

""" Sets up a CoAP server with a sample resource """

# Below implementation is based on the examples from:
# https://aiocoap.readthedocs.io/en/latest/examples.html

import argparse
import asyncio

from aiocoap import resource
import aiocoap


class SampleResource(resource.Resource):
    """
    Represents the resource stored on the server.
    Supports GET and PUT requests.

    Attributes
    ----------
    content : bytes
        represents the content stored on the server as bytes
    """

    def __init__(self, content: str):
        """
        Constructor to initialize the resource.

        Parameters
        ----------
        content : str
            content to initialize the resource with
        """
        # pylint: disable=non-parent-init-called
        super().__init__()
        self.set_content(content.encode("utf-8"))

    def set_content(self, content: bytes):
        """
        Set content for the resource. Encoded into
        bytes before setting.

        Parameters
        ----------
        content : bytes
            content to set for the resource
        """
        self.content = content

    # pylint: disable=unused-argument
    async def render_get(self, request):
        """
        Handler for GET requests received for the resource

        Parameters
        ----------
        request : aiocoap.Message
            GET request received from the client

        Returns
        -------
        aiocoap.Message
            response message for the GET request
        """
        return aiocoap.Message(payload=self.content)

    async def render_put(self, request):
        """
        Handler for PUT requests received for the resource

        Parameters
        ----------
        request : aiocoap.Message
            PUT request received from the client

        Returns
        -------
        aiocoap.Message
            response message for the PUT request
        """
        self.set_content(request.payload)
        # pylint: disable=no-member
        return aiocoap.Message(code=aiocoap.CHANGED, payload=self.content)


def main():
    """
    Parses the content for the server resource, sets up the
    root element, adds the resource to the root element, creates
    the server context and then runs the server process.
    """

    help_description = "CoAP server"

    # Initialize parser
    parser = argparse.ArgumentParser(description=help_description)

    # Adding arguments for client request
    parser.add_argument("-c", "--content", type=str)

    # Read Arguments
    args = parser.parse_args()

    # Resource tree creation, starting with the root element
    root_element = resource.Site()

    # For defining the entry point for requesting the list of links
    # for resources stored in the server. This is for performing
    # CoRE(Constrained RESTful Environments) Resource Discovery.
    root_element.add_resource(
        [".well-known", "core"],
        resource.WKCResource(root_element.get_resources_as_linkheader),
    )

    # Adding the SampleResource defined in this file
    root_element.add_resource(["sample-resource"], SampleResource(args.content))

    # Creating the server context
    asyncio.Task(aiocoap.Context.create_server_context(root_element))

    # This is a message for the parent process to consider this
    # background subprocess to have successfully run and to return an
    # exit code of 0
    print("Process successful\n")

    # Running the server
    asyncio.get_event_loop().run_forever()


if __name__ == "__main__":
    main()



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2022 NITK Surathkal

"""
Runs coap commands to setup CoAP experiment
and collect throughput data
"""

from functools import partial
import copy
import json
import logging

from nest.topology_map import TopologyMap
from .runnerbase import Runner
from ..results import CoAPResults
from ...engine.coap import run_coap_client, run_coap_server

logger = logging.getLogger(__name__)

# Certain unexpected errors occuring here are captured to this file.
log_file_handler = logging.FileHandler("coap_error.log")
log_file_handler.setLevel(logging.ERROR)
logger.addHandler(log_file_handler)


class CoAPRunner(Runner):
    """
    Runs coap client and server.

    Attributes
    ----------
    default_coap_client_options : dict
        Contains default options for the client to create messages
    user_options : dict
        Contains options provided by the user
    """

    # Contains the default CoAP client options
    # Will be used to form the CoAP client options string
    # to pass to the engine client file
    # More options can be added as per requirements
    default_coap_client_options = {
        "coap_message_payload": '-m "This is the default message for CoAP"',
        "coap_request_type": "-r get",
        "coap_non_timeout": "-t 5",
    }

    # pylint: disable=too-many-arguments
    def __init__(self, ns_id, destination_ip, user_options, n_con_msgs, n_non_msgs):
        """
        Constructor to initialize the runner

        Parameters
        ----------
        ns_id : str
            network namespace to run coap from
        destination_ip : str
            the ip of server to which it has to connect
        user_options: dict
            coap options provided by user to override default
        n_con_msgs : num
            number of CON messages to be sent
        n_non_msgs : num
            number of NON messages to be sent
        """
        self.user_options = user_options
        self.n_con_msgs = n_con_msgs
        self.n_non_msgs = n_non_msgs
        super().__init__(
            ns_id, start_time=None, run_time=None, destination_ip=destination_ip
        )

    @staticmethod
    def run_server(ns_id, server_options):
        """
        Run coap server in `ns_id`

        Parameters
        ----------
        ns_id : str
            namespace to run netserver on
        server_options : str
            options provided by the user for running the server
        """
        # Check if user has provided their own server options
        if server_options is not None:
            server_options_string = server_options

        # If not, use default options
        else:
            server_options_string = '-c "This is the default resource content for CoAP"'

        # Run CoAP server
        return_code = run_coap_server(ns_id, server_options_string)
        if return_code != 0:
            ns_name = TopologyMap.get_node(ns_id).name
            logger.error("Error running coap server at %s.", ns_name)

    def run(self):
        """
        Calls engine method to run coap client
        """
        coap_client_options = copy.copy(CoAPRunner.default_coap_client_options)

        # If the user has not supplied options, take the default options
        if self.user_options is None:
            coap_client_options.pop("coap_message_payload")
        else:
            # If user has provided the request type
            if "coap_request_type" in self.user_options.keys():
                # Convert the request type string to lowercase
                coap_request_type = self.user_options["coap_request_type"].lower()
            else:
                # If user has not provided the request type, take GET
                coap_request_type = "get"

            # Check validity of request type
            if coap_request_type not in ("get", "put"):
                logger.warning("Invalid request type %s.", coap_request_type)
                logger.warning("Taking 'get' as current request type.")
                coap_request_type = "get"

            coap_client_options["coap_request_type"] = f"-r {coap_request_type}"

            # Add message payload if 'put' request
            if coap_request_type == "put":
                if (
                    "coap_message_payload" in self.user_options.keys()
                    and self.user_options["coap_message_payload"] != ""
                ):
                    coap_message_payload = (
                        '"' + self.user_options["coap_message_payload"] + '"'
                    )
                    coap_client_options[
                        "coap_message_payload"
                    ] = f"-m {coap_message_payload}"

            # Message payload should not be sent for GET request
            else:
                coap_client_options.pop("coap_message_payload")

        # Add n_con_msgs and n_non_msgs
        coap_client_options["coap_n_con_msgs"] = f"-c {self.n_con_msgs}"
        coap_client_options["coap_n_non_msgs"] = f"-n {self.n_non_msgs}"

        # Add non_timeout to coap_client_options
        if "coap_non_timeout" in self.user_options.keys():
            non_timeout = self.user_options["coap_non_timeout"]
            coap_client_options["coap_non_timeout"] = f"-t {non_timeout}"

        # Convert coap_client_options dict to string for running engine
        coap_client_options_list = list(coap_client_options.values())
        coap_client_options_string = " ".join(coap_client_options_list)

        # Run CoAP client
        super().run(
            partial(
                run_coap_client,
                self.ns_id,
                self.destination_address.get_addr(with_subnet=False),
                self.destination_address.is_ipv6(),
                coap_client_options_string,
            ),
            error_string_prefix="Running coap",
        )

    def parse(self):
        """
        Parse coap output and error from self.out and self.err
        """
        # Rewind to start of the out file and read output
        self.out.seek(0)
        raw_stats_out = self.out.read().decode()

        # Rewind to start of the err file and read errors
        self.err.seek(0)
        raw_stats_err = self.err.read().decode()

        # Errors that occur outside the usual error
        # control systems are caught here. Since those
        # errors are still written to `stderr`, `json.loads`
        # would throw an exception for them not being in the
        # JSON format.
        #
        # Hence the exception is `JSONDecodeError`.
        try:
            # Parse outputs
            parsed_stats_out = {}
            if raw_stats_out != "":
                parsed_stats_out = json.loads(raw_stats_out)

            # Parse errors
            parsed_stats_err = {}
            if raw_stats_err != "":
                parsed_stats_err = json.loads(raw_stats_err)

        except json.decoder.JSONDecodeError:
            logger.error("Unexpected JSON Error. Error dump: %s", raw_stats_err)

        # Get Destination IP address and add results to runner
        destination_ip = self.destination_address.get_addr(with_subnet=False)
        CoAPResults.add_result(
            self.ns_id,
            {destination_ip: {"output": parsed_stats_out, "error": parsed_stats_err}},
        )



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2022 NITK Surathkal

"""
The base exception for all exceptions created in NeST
"""


class NestBaseException(Exception):
    """
    Base exception for all exceptions created in NeST
    """


class RequiredDependencyNotFound(Exception):
    """
    Raised when a depedency required for a experiment/topology
    is not installed.
    """

    def __init__(self, message="Depedency required for this program is not installed"):
        self.message = message
        super().__init__(self.message)


class DelayNotSetInInterface(NestBaseException):
    """
    Raised when delay is not set on the interface.
    """

    def __init__(
        self,
        message="Delay must be set on the interface before set_packet_reorder method is used."
        "\nPlease set delay first using set_delay or set_attributes methods",
    ):

        self.message = message
        super().__init__(self.message)


class DistributionOptionError(NestBaseException):
    """
    Raised when the delay distribution is not set properly
    """



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2024 NITK Surathkal

"""Plot MPEG-DASH results"""

import logging
import matplotlib.pyplot as plt
from ..interrupts import handle_keyboard_interrupt
from ..pack import Pack
from .common import simple_plot

logger = logging.getLogger(__name__)


def _get_list_of_plot_params():
    """
    Return list of params parsed by MPEG-DASH
    """
    return ["bitrate", "bitrate_level", "throughput", "buffer", "rtt"]


def _get_plot_params_units():
    """
    Obtain the units of each MPEG-DASH parameters
    """
    units = {
        "bitrate": "Kbps",
        "bitrate_level": "Kbps",
        "throughput": "Kbps",
        "buffer": "ms",
        "rtt": "s",
    }
    return units


def _get_ylabel(param):
    """
    Obtain y label for plots for a given parameter.

    For example, 'rtt' -> 'RTT (ms)'

    Arguments
    ---------
        param : str
            parameter

    Returns
    -------
        str
    """
    # Get MPEG-DASH param units
    units = _get_plot_params_units()
    # Get "nicer" param names
    nicer_param_names = {
        "bitrate": "Bitrate",
        "bitrate_level": "Bitrate Level",
        "throughput": "Throughput",
        "buffer": "Buffer",
        "rtt": "RTT",
    }

    ylabel = nicer_param_names[param]
    if units[param] is not None:
        ylabel += f" ({units[param]})"

    return ylabel


def _extract_from_mpeg_dash(flow, node, server_ip, stats_type):
    """
    Extract information from flow data and convert it to
    conveniently plottable data format

    Parameters
    ----------
    flow : List
        List with stats
    node : string
        The MPEG-DASH client node.
    server_ip : string
        IP Address of the MPEG-DASH server
    stats_type : string
        The type of stats for which plots are required. Can be 'Audio' or 'Video'.

    Returns
    -------
    flow_params or None
        Return flow parameters
    """
    # "meta" item will always be present, hence `<= 1`
    if len(flow) <= 1:
        logger.warning(
            "MPEG-DASH application from server %s to client %s doesn't have any parsed %s result.",
            server_ip,
            node,
            stats_type.lower(),
        )
        return None

    # First item is the "meta" item with user given information
    server_node = flow[0]["server_node"]

    chunk_numbers = []
    flow_params = {}

    for param in _get_list_of_plot_params():
        flow_params[param] = []
    for data in flow[1:]:
        for stat, stat_data in flow_params.items():
            if stat in data:
                stat_data.append(float(data[stat]))
            else:
                stat_data.append(None)
        chunk_numbers.append(data["chunk_number"])

    return {"server_node": server_node, "values": (chunk_numbers, flow_params)}


# pylint: disable=inconsistent-return-statements
def _plot_mpeg_dash(flow, node, server_ip, stats_type):
    """
    Plot MPEG-DASH stats

    Parameters
    ----------
    flow : List
        List with stats
    node : string
        The MPEG-DASH client node.
    server_ip : string
        IP Address of the MPEG-DASH server
    stats_type : string
        The type of stats for which plots are required. Can be 'Audio' or 'Video'.
    """
    data = _extract_from_mpeg_dash(flow, node, server_ip, stats_type)
    if data is None or data["values"] is None:
        return

    server_node = data["server_node"]
    (chunk_numbers, flow_params) = data["values"]

    legend_string = f"Server {server_node} to client {node}"
    for param in flow_params:
        fig = simple_plot(
            f"MPEG-DASH {stats_type} Statistics",
            chunk_numbers,
            flow_params[param],
            "Number of chunks",
            _get_ylabel(param),
            legend_string=legend_string,
        )

        filename = f"{stats_type.lower()}_{param}_{server_node}_to_{node}.png"
        Pack.dump_plot("mpeg_dash", filename, fig)
        plt.close(fig)
    return 0


# pylint: disable=too-many-locals
@handle_keyboard_interrupt
def plot_mpeg_dash(parsed_data):
    """
    Plot statistics obtained from MPEG-DASH

    Parameters
    ----------
    parsed_data : Dict
        JSON data parsed from MPEG-DASH application
    """

    # pylint: disable=too-many-nested-blocks
    for node in parsed_data:
        node_data = parsed_data[node]
        for connection in node_data:
            for server_ip in connection:
                flow_data = connection[server_ip]
                video_stats = flow_data["video"]
                try:
                    _plot_mpeg_dash(video_stats, node, server_ip, "Video")
                except Exception as excp:  # pylint: disable=broad-except
                    logger.error(
                        "Error generating plots for video statistics: %s", excp
                    )

                if flow_data["audio"][0]["audio_enabled"] is True:
                    audio_stats = flow_data["audio"]
                    try:
                        _plot_mpeg_dash(audio_stats, node, server_ip, "Audio")
                    except Exception as excp:  # pylint: disable=broad-except
                        logger.error(
                            "Error generating plots for audio statistics: %s", excp
                        )



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2024 NITK Surathkal

"""Custom HTTP Server"""
import http.server
import sys
import os


class CustomHttpHandler(http.server.SimpleHTTPRequestHandler):
    """
    CustomHTTPHandler that enables the server to serve a particular
    directory and also use HTTP/1.1 protocol.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, directory=sys.argv[2], **kwargs)


if __name__ == "__main__":
    if len(sys.argv) < 3:
        sys.exit(1)

    PORT = int(sys.argv[1])
    PATH = sys.argv[2]

    if not os.path.exists(PATH):
        sys.exit(1)

    http.server.test(
        HandlerClass=CustomHttpHandler,
        protocol="HTTP/1.1",
        port=PORT,
    )
